/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   fadis@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/

#include <k6x8/k6x8.h>
#include <k6x8/glyph.h>
#include <k6x8/map.h>
#include <k6x8/conv.h>
#include <k6x8/draw.h>

// フルスクリーンテキストレンダラを初期化する
void fstInit( FullScreenText *_fst, const char *_string ) {
  unsigned int index;
  for( index = 0; index != VIRTUAL_LINE_NUM; index++ )
    _fst->line_head[ index ] = NULL;
  _fst->buffer_head = 0;
  _fst->string_head = _string;
  _fst->scroll = 0;
  _fst->current_line = 0;
  _fst->end_line = VIRTUAL_LINE_NUM;
  _fst->mask = 0;
  _fst->displace = 0;
  _fst->displace_x_count = 0;
  _fst->displace_y_count = 0;
  for( index = 0; index != 3; index++ )
    fstNextLine( _fst );
}

// 内部リングバッファに次の1行を描画する
void fstNextLine( FullScreenText *_fst ) {
  if( _fst->line_head[ _fst->current_line ] ) {
    fstEUCJP( _fst->buffer + _fst->buffer_head * X_PIXEL_NUM, X_CHARCTER_NUM, _fst->line_head[ _fst->current_line ] );
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
  }
  else if( *( _fst->string_head ) ) {
    _fst->line_head[ _fst->current_line ] = _fst->string_head;
    _fst->string_head =
      fstEUCJP( _fst->buffer + _fst->buffer_head * X_PIXEL_NUM, X_CHARCTER_NUM, _fst->string_head );
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
  }
  else {
    _fst->line_head[ _fst->current_line ] = _fst->string_head;
    unsigned int col;
    for( col = 0; col != X_PIXEL_NUM; col++ )
      _fst->buffer [ _fst->buffer_head * X_PIXEL_NUM + col ] = 0;
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
    _fst->end_line = _fst->current_line;
  }
}

// 内部リングバッファに前の1行を描画する
void fstPreviousLine( FullScreenText *_fst ) {
  _fst->current_line--;
  _fst->buffer_head = ( _fst->buffer_head + 2 ) % 3;
  fstEUCJP( _fst->buffer + _fst->buffer_head * X_PIXEL_NUM, X_CHARCTER_NUM, _fst->line_head[ _fst->current_line - 2 ] );
}

// テキストをスクロールする　スクロールした結果内部バッファの描画済み範囲を越えていたら描画する
void fstScroll( FullScreenText *_fst, signed int _value ) {
  signed int scroll = _fst->scroll + _value;
  if( _value > 0 ) {
    for( ; scroll >= 8 && _fst->current_line < _fst->end_line; scroll -= 8 )
      fstNextLine( _fst );
    if( _fst->current_line < _fst->end_line )
      _fst->scroll = scroll & 0x07;
    else
      _fst->scroll = 0;
  }
  else if( _value < 0 ) {
    for( ; scroll < 0 && _fst->current_line > 2; scroll += 8 )
      fstPreviousLine( _fst );
    if( _fst->current_line > 2 )
      _fst->scroll = (unsigned int)scroll & 0x07;
    else
      _fst->scroll = 0;
  }
}

// テキストが終端に達していたら非ゼロ値を返す
int fstIsEnd( FullScreenText *_fst ) {
  return _fst->current_line == _fst->end_line;
}

// フェードイン/フェードアウト用のマスク値
// もう少し容量をケチることも出来るが、描画時のパフォーマンスに影響するのでやめておく
const uint8_t fst_fade_mask[ 2 * 5 ] = {
  // ■■
  // ■■
  0xFF, 0xFF,
  // ■■
  // ■□
  0xFF, 0xAA,
  // ■□
  // □■
  0xAA, 0x55,
  // ■□
  // □□
  0xAA, 0x00,
  // □□
  // □□
  0x00, 0x00,
};

// 振動の素
const uint32_t fst_displace_random_x = 0x18FD335E;
const uint32_t fst_displace_random_y = 0x957E1899;

// 内部バッファとスクロール値から画面に描画するイメージを作る
void fstMakeScrolledImage( FullScreenText *_fst ) {
  unsigned int dest_line;
  for( dest_line = 0; dest_line != 2; dest_line++ ) {
    unsigned int source_line0 = ( _fst->current_line + dest_line ) % 3;
    unsigned int source_line1 = ( _fst->current_line + dest_line +  1 ) % 3;
    unsigned int charcter;
    for( charcter = 0; charcter != X_CHARCTER_NUM; charcter++ ) {
      unsigned int col;
      unsigned int current_displace_x;
      unsigned int current_displace_y;
      if( _fst->displace ) { // 文字を振動させる
        current_displace_x = ( fst_displace_random_x >> _fst->displace_x_count ) & 0x01;
        current_displace_y = ( fst_displace_random_y >> _fst->displace_y_count ) & 0x01;
        // XとYの周期の最小公倍数を大きくすることで複雑そうに見える振動を作る
        _fst->displace_x_count++;
        _fst->displace_x_count %= 31;
        _fst->displace_y_count++;
        _fst->displace_y_count %= 30;
      }
      else { // 文字を振動させない
        current_displace_x = 0;
        current_displace_y = 0;
      }
      for( col = 0; col != 6; col++ ) {
        _fst->scrolled_buffer[ charcter * 6 + col + dest_line * X_PIXEL_NUM ] =
          ( ( ( _fst->buffer[ source_line0 * X_PIXEL_NUM + charcter * 6 + ( ( 6 + col - current_displace_x ) % 6 ) ] >> _fst->scroll ) |
          ( _fst->buffer[ source_line1 * X_PIXEL_NUM + charcter * 6 + ( ( 6 + col - current_displace_x ) % 6 ) ] << ( 8 - _fst->scroll ) ) ) << current_displace_y )
          & fst_fade_mask[ _fst->mask * 2 + ( col & 0x01 ) ];
      }
    }
  }
}

// フェードイン/フェードアウト用マスクをセットする 以後描画される文字には選択したマスクが適用される
void fstSetMask( FullScreenText *_fst, uint8_t _level ) {
  _fst->mask = _level;
}

// 文字を振動させる _valueが非ゼロなら次の描画から文字が振動する ゼロなら振動しない
void fstSetDisplace( FullScreenText *_fst, uint8_t _value ) {
  _fst->displace = _value;
}

// 振動とスクロールを適用した結果の画像を得る
// このライブラリの中で描画関数を呼び出してしまうと
// スタティックライブラリにOLEDドライバを含んでしまい サイズとライセンスの両面で面倒なので
// 最終的な描画はユーザプログラム側で行う
const uint8_t *fstGetBuffer( FullScreenText *_fst ) {
  fstMakeScrolledImage( _fst );
  return _fst->scrolled_buffer;
}
